﻿#ZLIB_VERSION = "1.2.8"
#WANT_GZIP = 16

#Z_BUF_ERROR     = -5
#Z_FINISH        =  4

#ENABLE_ZLIB_GZIP = 32
#Z_DEFAULT_COMPRESSION = -1
#Z_DEFLATED = 8
#MAX_MEM_LEVEL = 9
#Z_DEFAULT_STRATEGY = 0
#Z_NULL = 0
#OS_NTFS = 11



#dodebug=#False

Structure Z_STREAM Align #PB_Structure_AlignC
  *next_in.Byte
  avail_in.l
  total_in.l
  *next_out.Byte
  avail_out.l
  total_out.l
  *msg.Byte
  *state
  zalloc.l
  zfree.l
  opaque.l
  data_type.i
  adler.l
  reserved.l
  CompilerIf #PB_Compiler_Processor = #PB_Processor_x64
    alignment.l
  CompilerEndIf
EndStructure

ImportC "zlib.lib"
  inflateInit2_(*strm, windowBits.i, version.s, strm_size)
  inflate(*strm, flush.i)
  inflateEnd(*strm)
  
  deflateInit2_(*strm, level.i, method.i, windowBits.i, memlevel.i, strategy.i, Version.s, strm_size)
  deflateBound(*strm, sourceLen)
  deflate(*strm, flush.i)
  deflateEnd(*strm)
  
EndImport

Procedure.l zlib_inflateInit2(*dest, *destLen.long, *source, sourceLen.l)
   strm.Z_STREAM
  FillMemory(@strm,SizeOf(strm))
  
  With strm
    \zalloc = #Null
    \zfree  = #Null
    \opaque = #Null
    
    \next_in  = *source
    \avail_in = sourceLen
    \total_in = #Null
    
    \next_out   = *dest
    \avail_out  = *destLen\l
    \total_out  = #Null   
  EndWith
  
  err.l = inflateInit2_(@strm, #WANT_GZIP, #ZLIB_VERSION, SizeOf(Z_STREAM))
  If err <> 0 : ProcedureReturn err : EndIf
  
  err = inflate(@strm,#Z_FINISH)
  
  If err <> 1
    inflateEnd(@strm)
    If err = 0
      ProcedureReturn #Z_BUF_ERROR
    Else
      ProcedureReturn err
    EndIf
  EndIf
  *destlen\l=strm\total_out
  ProcedureReturn inflateEnd(@strm) 
EndProcedure
Procedure.l zlib_deflateInit2(*dest.quad,*destLen.long, *source,SourceLen.l)
  strm.Z_STREAM
  strm\next_in=*source
  strm\avail_in=sourcelen
  
  If *dest\q
    FreeMemory(*dest\q)
  EndIf
  *dest\q=0
  *destlen\l=0
  
  err= deflateInit2_(@strm,#Z_DEFAULT_COMPRESSION,#z_Deflated,15|16,#MAX_MEM_LEVEL,#Z_DEFAULT_STRATEGY,#ZLIB_VERSION,SizeOf(z_stream))
  ;debug "deflate "+Str(err)
  If err<>0:ProcedureReturn 0:EndIf
  *destlen\l=deflateBound(@strm,sourcelen)
  *dest\q=AllocateMemory(*destlen\l)
  strm\next_out=*dest\q
  strm\avail_out=*destlen\l
  
  err=Deflate(@strm, #Z_FINISH)
  *destlen\l=strm\total_out
  deflateEnd(@strm)
  
EndProcedure


Procedure.s ExplainIt(val)
  Str.s
  If val = 0 : Str = "#Z_OK" : EndIf
  If val = 1 : Str = "#Z_STREAM_END" : EndIf
  If val = 2 : Str = "#Z_NEED_DICT" : EndIf
  If val = -1 : Str = "#Z_ERRNO" : EndIf
  If val = -2 : Str = "#Z_STREAM_ERROR" : EndIf
  If val = -3 : Str = "#Z_DATA_ERROR" : EndIf
  If val = -4 : Str = "#Z_MEM_ERROR" : EndIf
  If val = -5 : Str = "#Z_BUF_ERROR" : EndIf
  If val = -6 : Str = "#Z_VERSION_ERROR" : EndIf
  ProcedureReturn Str
EndProcedure

Enumeration
  #TAG_End 
  #TAG_Byte 
  #TAG_Short
  #TAG_Int
  #TAG_Long
  #TAG_Float 	
  #TAG_Double 	
  #TAG_Byte_Array
  #TAG_String 
  #TAG_List 	
  #TAG_Compound
  #TAG_Int_Array
EndEnumeration

Structure NTB
  tag.i
  len.l
  *pos
EndStructure

Global NewMap ntb.ntb()

Procedure.q NTB_GetValue(key$)
  ret.q=0
  If ntb(key$)\pos >0
    *byte.byte=ntb(key$)\pos
    Select ntb(key$)\tag
      Case #TAG_Byte
        ret=*byte\B &$ff:*byte+1
      Case #TAG_Short
        ret=(*byte\B &$ff)<<8:*byte+1
        ret+(*byte\B &$ff):*byte+1
      Case #TAG_int
        ret=(*byte\B &$ff)<<24:*byte+1
        ret+(*byte\B &$ff)<<16:*byte+1
        ret+(*byte\B &$ff)<<8:*byte+1
        ret+(*byte\B &$ff):*byte+1
      Case #TAG_Long
        ret=(*byte\B &$ff)<<56:*byte+1
        ret+(*byte\B &$ff)<<48:*byte+1
        ret+(*byte\B &$ff)<<40:*byte+1
        ret+(*byte\B &$ff)<<32:*byte+1
        ret+(*byte\B &$ff)<<24:*byte+1
        ret+(*byte\B &$ff)<<16:*byte+1
        ret+(*byte\B &$ff)<<8:*byte+1
        ret+(*byte\B &$ff):*byte+1
      Default
        Debug "#### UNKNOWN TAG "+key$
    EndSelect    
  Else
    Debug "#### UNKNWON KEY "+key$
  EndIf
  ProcedureReturn ret  
EndProcedure
Procedure.q NTB_SetValue(key$,q.q)
  ret.q=0
  If ntb(key$)\pos>0
    *byte.byte=ntb(key$)\pos
    Select ntb(key$)\tag
      Case #TAG_Byte
        ;debug "setbyte"
        *byte\b=q &$ff
      Case #tag_short
        ;debug "setshort"
        *byte\b=(q>>8)&$ff:*byte+1
        *byte\b=q&$ff
      Case #TAG_Int
        ;debug "setint"
        *byte\b=(q>>24)&$ff:*byte+1
        *byte\b=(q>>16)&$ff:*byte+1
        *byte\b=(q>>8)&$ff:*byte+1
        *byte\b=q&$ff
      Case #tag_long
        ;debug "setlong"
        *byte\b=(q>>56)&$ff:*byte+1
        *byte\b=(q>>48)&$ff:*byte+1
        *byte\b=(q>>40)&$ff:*byte+1
        *byte\b=(q>>32)&$ff:*byte+1
        *byte\b=(q>>24)&$ff:*byte+1
        *byte\b=(q>>16)&$ff:*byte+1
        *byte\b=(q>>8)&$ff:*byte+1
        *byte\b=q&$ff
    EndSelect
  EndIf
EndProcedure


Procedure.d NTB_GetValueD(key$)
  ret.d=0
  If ntb(key$)\pos >0 And ntb(key$)\tag=#TAG_Double
    *byte.byte=ntb(key$)\pos
    q.q=(*byte\B &$ff)<<56:*byte+1
    q+(*byte\B &$ff)<<48:*byte+1
    q+(*byte\B &$ff)<<40:*byte+1
    q+(*byte\B &$ff)<<32:*byte+1
    q+(*byte\B &$ff)<<24:*byte+1
    q+(*byte\B &$ff)<<16:*byte+1
    q+(*byte\B &$ff)<<8:*byte+1
    q+(*byte\B &$ff):*byte+1
    ret.d=PeekD(@q)
  EndIf
  ProcedureReturn ret
EndProcedure
Procedure.f NTB_GetValueF(key$)
  ret.f=0
  If ntb(key$)\pos >0 And ntb(key$)\tag=#TAG_Float
    *byte.byte=ntb(key$)\pos
    i=(*byte\B &$ff)<<24:*byte+1
    i+(*byte\B &$ff)<<16:*byte+1
    i+(*byte\B &$ff)<<8:*byte+1
    i+(*byte\B &$ff):*byte+1
    ret.f=PeekF(@i)
  EndIf
  ProcedureReturn ret
EndProcedure
Procedure.s NTB_GetValueS(key$)
  ret$=""
  If ntb(key$)\pos >0 And ntb(key$)\tag=#TAG_String
    *byte.byte=ntb(key$)\pos
    ret$=PeekS(ntb(key$)\pos,ntb(key$)\len,#PB_UTF8)
  EndIf
  ProcedureReturn ret$
EndProcedure



Global _ntb_size=0,_ntb_datasize=0
Global *_ntb=0
Procedure FreeNTB()
  If *_ntb
    FreeMemory(*_ntb)
  EndIf
  _ntb_size=0
  *_ntb=0
  ClearMap(ntb())
EndProcedure
Procedure SaveNTB(file$)
  If *_ntb And _ntb_size
    *mem=0
    memsize=0
    ;debug "save "+Hex(*_ntb)+" "+Str(_ntb_datasize)
    zlib_deflateInit2(@*mem,@memsize,*_ntb,_ntb_datasize)
    ;debug memsize
    ;debug Hex(*mem)
    If *mem
      CreateFile(1,file$)
      WriteData(1,*mem,memsize)
      CloseFile(1)
      FreeMemory(*mem)
    EndIf
    
    
  EndIf
  
EndProcedure

Structure taglist
  len.l
  count.l
  deep.l
  tag.l
  name$
EndStructure


Procedure LoadNTB(file$,dounpack=#True)
  ClearMap(ntb())
  NewList taglist.taglist()
  
  
  If dounpack 
    ReadOK = ReadFile(#PB_Any, file$)
    If ReadOK
      CompressedLength = Lof(ReadOK) : 
      *Compressed = AllocateMemory(CompressedLength)
      If *compressed=0
        CloseFile(readok)
        ProcedureReturn #False
      EndIf
      ReadData(ReadOK, *Compressed, CompressedLength)
      CloseFile(readok)
    Else
      ProcedureReturn #False
    EndIf
    
    
    UnCompressedLength = CompressedLength * 10
    
    If UnCompressedLength=0
      ProcedureReturn #False
    EndIf
    
    
    If _ntb_size<UnCompressedLength Or _ntb_size>UnCompressedLength+1024
      ;debug Hex(*_ntb)+"  "+Str(_ntb_size)+"  "+Str(UnCompressedLength)
      If *_ntb
        freentb()
      EndIf
      
      ;  *_ntb=ReAllocateMemory(*_ntb,UnCompressedLength)
      ;Else
      *_ntb=AllocateMemory(UnCompressedLength+1024)
      ;EndIf
      _ntb_size=UnCompressedLength+10
      ;debug Hex(*_ntb)+"  memory"
    EndIf
    
    
    If *_ntb=0
      FreeMemory(*Compressed)
      ProcedureReturn #False
    EndIf
    
    FillMemory(*_ntb,_ntb_size)
    
    *UnCompressed = *_ntb
    
    Z_Error.b = zlib_inflateInit2(*UnCompressed, @UnCompressedLength, *Compressed, CompressedLength)
    FreeMemory(*Compressed)
    
  Else
    ReadOK = ReadFile(#PB_Any, file$)
    If ReadOK
      UnCompressedLength = Lof(ReadOK) : 
      *uncompressed = AllocateMemory(UnCompressedLength+10)
      If *uncompressed=0
        CloseFile(readok)
        ProcedureReturn #False
      EndIf
      ReadData(ReadOK, *uncompressed, UnCompressedLength)
      CloseFile(readok)
      z_error=0
    Else
      ProcedureReturn #False
    EndIf
  EndIf

  
  
  If Z_Error=0
    ;Debug "unpack:"+Str(UnCompressedLength)
    _ntb_datasize=UnCompressedLength
    *byte.byte=*uncompressed
    *byteend=*UnCompressed+UnCompressedLength
    ein$="" 
    
    ;listlen=0:deep=0
    AddElement (taglist())
    
    
    
    Repeat
      If taglist()\len=0 Or taglist()\deep>0
        tag=*byte\b &$ff:*byte+1
      Else
        tag=taglist()\tag
      EndIf
      
      
      If tag<>#TAG_End
        
        If taglist()\len=0
          slen=(*byte\b &$ff)<<8:*byte+1
          slen+(*byte\b &$ff):*byte+1
          key$=PeekS(*byte,slen):*byte+slen
        Else
          
          
          If taglist()\deep=0
            key$=taglist()\name$+"."+Str(taglist()\count)
          Else              
            slen=(*byte\b &$ff)<<8:*byte+1
            slen+(*byte\b &$ff):*byte+1
            key$=PeekS(*byte,slen):*byte+slen
          EndIf
          
        EndIf
        
        
        Select tag
          Case #TAG_Byte
            ntb(ein$+key$)\len=1
            ntb(ein$+key$)\pos=*byte
            ntb(ein$+key$)\tag=tag
            CompilerIf #dodebug
              b=*byte\B &$ff:*byte+1
              Debug ein$+key$ +" byte "+Str(b):
            CompilerElse
              *byte+1
            CompilerEndIf
            
          Case #TAG_Short
            ntb(ein$+key$)\len=2
            ntb(ein$+key$)\pos=*byte
            ntb(ein$+key$)\tag=tag
            CompilerIf #dodebug
              s=(*byte\B &$ff)<<8:*byte+1
              s+(*byte\B &$ff):*byte+1
              Debug ein$+key$+" short "+Str(s)
            CompilerElse
              *byte+2
            CompilerEndIf
            
          Case #TAG_int
            ntb(ein$+key$)\len=4
            ntb(ein$+key$)\pos=*byte
            ntb(ein$+key$)\tag=tag
            CompilerIf #dodebug
              i=(*byte\B &$ff)<<24:*byte+1
              i+(*byte\B &$ff)<<16:*byte+1
              i+(*byte\B &$ff)<<8:*byte+1
              i+(*byte\B &$ff):*byte+1
              Debug ein$+key$+" Int "+Str(i)
            CompilerElse
              *byte+4
            CompilerEndIf
            
          Case #TAG_Long
            ntb(ein$+key$)\len=8
            ntb(ein$+key$)\pos=*byte
            ntb(ein$+key$)\tag=tag
            CompilerIf #dodebug
              q.q=(*byte\B &$ff)<<56:*byte+1
              q+(*byte\B &$ff)<<48:*byte+1
              q+(*byte\B &$ff)<<40:*byte+1
              q+(*byte\B &$ff)<<32:*byte+1
              q+(*byte\B &$ff)<<24:*byte+1
              q+(*byte\B &$ff)<<16:*byte+1
              q+(*byte\B &$ff)<<8:*byte+1
              q+(*byte\B &$ff):*byte+1
              Debug ein$+key$+" Long "+Str(q)
            CompilerElse
              *byte+8
            CompilerEndIf
            
          Case #TAG_Float
            ntb(ein$+key$)\len=4
            ntb(ein$+key$)\pos=*byte
            ntb(ein$+key$)\tag=tag
            CompilerIf #dodebug
              i=(*byte\B &$ff)<<24:*byte+1
              i+(*byte\B &$ff)<<16:*byte+1
              i+(*byte\B &$ff)<<8:*byte+1
              i+(*byte\B &$ff):*byte+1
              f.f=PeekF(@i)
              Debug ein$+key$+" float "+StrF(f)
            CompilerElse
              *byte+4
            CompilerEndIf
            
          Case #TAG_Double
            ntb(ein$+key$)\len=8
            ntb(ein$+key$)\pos=*byte
            ntb(ein$+key$)\tag=tag
            CompilerIf #dodebug
              q.q=(*byte\B &$ff)<<56:*byte+1
              q+(*byte\B &$ff)<<48:*byte+1
              q+(*byte\B &$ff)<<40:*byte+1
              q+(*byte\B &$ff)<<32:*byte+1
              q+(*byte\B &$ff)<<24:*byte+1
              q+(*byte\B &$ff)<<16:*byte+1
              q+(*byte\B &$ff)<<8:*byte+1
              q+(*byte\B &$ff):*byte+1
              d.d=PeekD(@q)
              Debug ein$+key$+" double "+StrD(d)
            CompilerElse
              *byte+8
            CompilerEndIf
            
          Case #TAG_Byte_Array
            i=(*byte\B &$ff)<<24:*byte+1
            i+(*byte\B &$ff)<<16:*byte+1
            i+(*byte\B &$ff)<<8:*byte+1
            i+(*byte\B &$ff):*byte+1
            
            ntb(ein$+key$)\len=i
            ntb(ein$+key$)\pos=*byte
            ntb(ein$+key$)\tag=tag
            
            CompilerIf #dodebug
              a$=""
              For x=1 To i
                If x<10
                  a$+Str(*byte\b &$ff)+" "
                EndIf
                *byte+1
              Next
              
              Debug ein$+key$+ " Byte_Array (anzahl:"+Str(i)+") "+a$
            CompilerElse
              *byte+i
            CompilerEndIf
            
            
          Case #TAG_String
            slen=(*byte\b &$ff)<<8:*byte+1
            slen+(*byte\b &$ff):*byte+1
            
            ntb(ein$+key$)\len=slen
            ntb(ein$+key$)\pos=*byte
            ntb(ein$+key$)\tag=tag
            
            CompilerIf #dodebug
              s$=PeekS(*byte,slen,#PB_UTF8):*byte+slen
              Debug ein$+key$+ " string "+s$
            CompilerElse
              *byte+slen
            CompilerEndIf
            
          Case #tag_list  
            AddElement(taglist())
            
            taglist()\tag=*byte\b:*byte+1
            i=(*byte\B &$ff)<<24:*byte+1
            i+(*byte\B &$ff)<<16:*byte+1
            i+(*byte\B &$ff)<<8:*byte+1
            i+(*byte\B &$ff):*byte+1
            taglist()\len=i+1;Korrektur
            CompilerIf #dodebug
              Debug "List ("+key$+") Anzahl:"+Str(i)
            CompilerEndIf
            
            taglist()\name$=key$:taglist()\count=0
            
          Case #TAG_Compound
            CompilerIf #dodebug
              Debug "#cmpound ("+key$+")"
            CompilerEndIf
            ein$+key$+"\"
            If taglist()\len
              taglist()\deep+1
            EndIf
            
          Case #TAG_Int_Array
            i=(*byte\B &$ff)<<24:*byte+1
            i+(*byte\B &$ff)<<16:*byte+1
            i+(*byte\B &$ff)<<8:*byte+1
            i+(*byte\B &$ff):*byte+1
            
            ntb(ein$+key$)\len=i
            ntb(ein$+key$)\pos=*byte
            ntb(ein$+key$)\tag=tag
            
            *byte+(i*2)
            CompilerIf #dodebug
              Debug ein$+key$+ " Byte_Array (anzahl:"+Str(i)+") "
            CompilerEndIf
            
          Default
            Debug ein$+ "UNKNOWN "+ein$+key$+" "+Str(tag)+" "+lastkey$+" "+Str(lasttag)
            Debug Hex(*byte)+" "+Hex(*byteend)
            *byte=*byteend+1
        EndSelect
        
        lasttag=0
        lastkey$=ein$+key$
        
      Else
        ein$=GetPathPart(ein$);Left(ein$,Len(ein$)-2)
        a=0:c=0
        Repeat
          b=FindString(ein$,"\",a+1)
          If b
            c=a
            a=b
          Else
            Break
          EndIf
        ForEver
        ein$=Left(ein$,c)
        CompilerIf #dodebug
          Debug "#END"
        CompilerEndIf
        If taglist()\len
          taglist()\deep-1
        EndIf
        
      EndIf
      
      
      If taglist()\len And taglist()\deep=0
        taglist()\len-1
        taglist()\count+1
        If taglist()\len=0
          DeleteElement(taglist())
        EndIf  
      EndIf
    Until *byte>*byteend
    CompilerIf #dodebug
      Debug Hex(*byte)+" "+Hex(*byteend)
    CompilerEndIf    
    ret=#True
    
    
  Else :
    ret=#False
    Debug "zlib_inflateInit2_(): Error "+ExplainIt(Z_Error)
  EndIf   
  
  
  ProcedureReturn ret
EndProcedure

Procedure debugntb()
  ForEach ntb()
    Debug MapKey(ntb())
  Next
EndProcedure



CompilerIf #dodebug
  file$="E:\Downloads\bigtest.nbt"
  ;loadntb(file$)
  ;freentb()
  file$="F:\minecraft\gamedir\Aktueller Release\saves\Demowelt\level.dat"
  file$="F:\minecraft\gamedir\Aktueller Snapshot\saves\Hometown\level.dat"
If LoadNTB(file$)
  Debug "loaded"
  Debug NTB_GetValueS("\Data\LevelName")
  Debug ntb_getvalue("\Data\LastPlayed")
  Debug FormatDate("%dd.%mm.%yyyy %hh:%ii",ntb_getvalue("\Data\LastPlayed")/1000)
  
  Debug NTB_GetValue("\Data\Player\playerGameType")
  Debug ntb_Getvalue("\Data\RandomSeed")
  Debug ntb_getvalue("\Data\GameType")
  Debug NTB_GetValue("\Data\MapFeatures")
  Debug NTB_GetValue("\Data\allowCommands")
  Debug ntb_getvalues("\Data\generatorName")
  ;ntb_setvalue("\Data\GameType",1)
  ;ntb_setvalue("\Data\Player\playerGameType",2)
  Debug ntb_getvalue("\Data\GameType")
  Debug NTB_GetValue("\Data\Player\playerGameType")
  ;ntb_setvalue("\Data\hardcore",0)
  ;ntb_setvalue("\Data\DifficultyLocked",0)
  ;ntb_setvalue("\Data\Difficulty",2)
  Debug "test"
  ;ntb_setvalue("\Data\Player\abilities\mayfly",1)
  ;ntb_setvalue("\Data\Player\abilities\instabuild",1)
  ;ntb_setvalue("\Data\Player\Invulnerable",1);unsterblich
  ;NTB_setValue("\Data\allowCommands",1)
  ;SaveNTB(file$)
  
EndIf
End
CompilerEndIf

; IDE Options = PureBasic 5.30 (Windows - x64)
; CursorPosition = 365
; FirstLine = 361
; Folding = -+----
; EnableXP
; EnablePurifier